/**!
 * Copyright (c) All rights reserved.
 *
 * @author Dheeraj Kumar Aggarwal <dheeraj.aggarwal@optimizory.com>
 *
 * testRunner.js : Test Runner to execute a single HTTP request
 */

var savedRequestHeaders = {},
    //unsafe headers list - http://www.w3.org/TR/XMLHttpRequest/#the-setrequestheader%28%29-method
    unsafeHeaders = [ "accept-charset", "accept-encoding", "access-control-request-headers",
                      "access-control-request-method", "connection", "content-length", "cookie",
                      "cookie2", "date", "dnt", "expect", "host", "keep-alive", "origin", "referer",
                      "te", "trailer", "transfer-encoding", "upgrade", "user-agent", "via"];

var url_encode = function(params){
  var encoded = '', i, count;
  for(i = 0, count = params.length; i < count; i++){
    if(params[i].method == 'body' && params[i].name.length) {
      if(encoded.length)  encoded += '&';
      encoded += encodeURIComponent(params[i].name) + '=' + encodeURIComponent(params[i].value);
    }
  }
  return encoded;
};

var TC_CALLBACK = {},
  TC_REQ_ID = {},
  CURRENT_XHR = null,
  REQ_CALLBACK = {};


var isNumber = function(n) {
  return !isNaN(parseFloat(n)) && isFinite(n);
};

var parseBoolean = function(val) {
  if(val === 'true') return true;
  else if(val === 'false') return false;
  else return val;
};

var parseObject = function(val) {
  try {
    return JSON.parse(val);
  } catch(e) {
    return val;
  }
};

var json_encode = function(params){
  var encoded = {};
  for(i = 0, count = params.length; i < count; i++){
    var key = params[i].name;
    var val = params[i].value;
    if(params[i].method == 'body' && key.length) {
      var type = params[i].paramType;
      if(type == 'number') encoded[key] = isNumber(val) ? parseFloat(val) : val;
      else if(type == 'boolean') encoded[key] = parseBoolean(val);
      else if(type == 'object') encoded[key] = parseObject(val);
      else encoded[key] = val;
    }
  }
  return JSON.stringify(encoded);
};

var parseResponseHeaders = function(response){
  var headers = {}, i, count;
  if(response && response.responseHeaders){
    var pairs = response.responseHeaders.split("\r\n"), str, pair;
    for(i = 0, count = pairs.length; i < count; i++){
      str = pairs[i];
      if(str){
        pair = str.split(": ");
        if(pair && pair.length === 2){
          headers[pair[0]] = pair[1];
        }
      }
    }
  }
  return headers;
};

var runTestCase = function(oArgs, callback){
  var tc = oArgs.testcase;

  var i, count, method = tc.method,
    url = tc.url,
    headers = {},
    saveHeaders = {},
    params = tc.params;

  if(tc.headers && tc.headers.length > 0){
    for(i = 0, count = tc.headers.length; i < count; i++){
      if(tc.headers[i].name){
        //Got error Refused to set unsafe header "User-Agent", if I comment the below line, but no error in sending request
        if(unsafeHeaders.indexOf(tc.headers[i].name.toLowerCase()) === -1){
          headers[tc.headers[i].name] = tc.headers[i].value;
        } else {
          saveHeaders[tc.headers[i].name] = tc.headers[i].value;
        }
      }
    }
  }

  if(tc.authorizationHeader){
    headers["authorization"] = tc.authorizationHeader;
  }

  headers['x-powered-by'] = 'vREST';
  headers['x-vrest-testcaseId'] = tc.id;

  var options = {
    url: url,
    method: method,
    headers: headers
  };
  if(typeof oArgs.timeout === 'number'){
    options.timeout = oArgs.timeout;
  }
  var body = null, xbody;
  if(tc.raw && tc.raw.content && tc.raw.enabled === true){
    xbody = tc.raw.content;
  }
  if((['POST', 'PATCH', 'PUT', 'DELETE'].indexOf(method) != -1)
    && ((params && params.length > 0) || xbody)) {
    var contentType = null;
    if(headers['Content-Type'] || headers['content-type'])
      contentType = headers['Content-Type'] || headers['content-type'];

    if(xbody) {
      body = xbody;
      if(!contentType){
        try {
          JSON.parse(body);
          options.headers['Content-Type'] = 'application/json';
        } catch(e){} }
    }
    else if(contentType && contentType.indexOf("json") != -1)
      body = json_encode(tc.params);
    else {
      options.headers['Content-Type'] = 'application/x-www-form-urlencoded';
      body = url_encode(tc.params);
    }
  }

  for(var key in options.headers){
    saveHeaders[key] = options.headers[key];
  }

  savedRequestHeaders[tc.id] = saveHeaders;

  if(body){
    options.body = body;
  }
  var currentCB = function(err, response){
    CURRENT_XHR = null;
    var result = {
      body: response.body,
      responseHeaders: parseResponseHeaders(response),
      responseText: response.responseText,
      status: response.status,
      statusCode: response.statusCode
    };

    if(TC_REQ_ID[tc.id]){ //cleanup
      delete REQ_CALLBACK[TC_REQ_ID[tc.id]];
      delete TC_REQ_ID[tc.id];
    }

    if(savedRequestHeaders[tc.id]){
      try {
        options.headers = JSON.parse(JSON.stringify(savedRequestHeaders[tc.id]));
      } catch(e){
      }
    }
    delete savedRequestHeaders[tc.id];

    callback({err: err, response: result, runnerCase : options});
  };

  TC_CALLBACK[tc.id] = currentCB;
  CURRENT_XHR = request(options, currentCB);
  return true;
};

chrome.webRequest.onErrorOccurred.addListener(
  function(details){
    var callback = REQ_CALLBACK[details.requestId];

    if(callback){
      if(details.error === 'net::ERR_NAME_NOT_RESOLVED'){
        callback(details.error + ' - Domain name of request URL not resolved.', {});
      } else if(details.error === 'net::ERR_INSECURE_RESPONSE'){
        callback(details.error + ' - Error occurred probably due to the use of self signed SSL certificate. ' +
          'To resolve this issue, please open the base URL in another tab and accept the certificate.', {});
      } else callback(details.error, {});
    }
  },
  {urls: ["<all_urls>"]});

chrome.runtime.onMessage.addListener(function(oArgs, sender, callback) {
  if(oArgs.message){
    if(oArgs.message === 'version'){
      var manifest = chrome.runtime.getManifest();
      return callback({version: manifest.version});
    } else if(oArgs.message === 'testcase'){
      return runTestCase(oArgs, callback);
    } else if(oArgs.message === 'abortrequest'){
      if(CURRENT_XHR) CURRENT_XHR.abort();
      return;
    } else if(oArgs.message === 'callrest'){
      return request(oArgs.data, function(err, data, json){
        if(err && err.name === 'SyntaxError' && data
            && (data.status === 200 || data.status === 201)){
          err = null;
          json = data = data.responseText || 1;
        }
        if(err || !json || !data) {
          callback({
            error : (err && err.message) || err || 'No output present'
          });
        } else {
          if (typeof json === 'object' && json) {
            if (Object.keys(json) === 1) {
              return callback({ output : json.output || json || data });
            }
          }
          return callback(json || data);
        }
      });
    } else {
      return callback({ err : 'Invalid message request.' });
    }
  }
  return true;
});

var getHeaderValue = function(headerName, headers){
  var i, count;
  for(i = 0, count = headers.length; i < count; i++){
    if(headers[i].name === headerName){
      return headers[i].value;
    }
  }
  return null;
};

chrome.webRequest.onBeforeSendHeaders.addListener(
  function(details) {
    var testCaseId = getHeaderValue('x-vrest-testcaseId', details.requestHeaders),
      poweredBy = getHeaderValue('x-powered-by', details.requestHeaders);

    if(poweredBy && poweredBy === "vREST" && testCaseId && savedRequestHeaders[testCaseId]){
      var savedHeaders = savedRequestHeaders[testCaseId];

      /* Set Test Case Callback with request id */
      TC_REQ_ID[testCaseId] = details.requestId;
      REQ_CALLBACK[details.requestId] = TC_CALLBACK[testCaseId];
      delete TC_CALLBACK[testCaseId];

      var headers = [];
      if(savedHeaders){
        var i, count;
        for(i = 0, count = details.requestHeaders.length; i < count; i++){
          if(savedHeaders[details.requestHeaders[i].name]){
            details.requestHeaders[i].value = savedHeaders[details.requestHeaders[i].name];
            delete savedHeaders[details.requestHeaders[i].name];
          }
        }

        for(var key in savedHeaders){
          details.requestHeaders.push({
            name: key,
            value: savedHeaders[key]
          });
        }

        for(i = 0, savedRequestHeaders[testCaseId] = {}, count = details.requestHeaders.length; i < count; i++){
          savedRequestHeaders[testCaseId][details.requestHeaders[i].name] = details.requestHeaders[i].value;
        }
      }

    }

    return {requestHeaders: details.requestHeaders};
  },
  {urls: ["<all_urls>"]},
  ["blocking", "requestHeaders"]
);
